@@ -1,3 +1,4 @@
+import collections
 import datetime
 import functools
 import gzip
@@ -16,10 +17,16 @@ from collections import defaultdict
 import click
 from tqdm import tqdm
 
+# size of buffer to use when reading files for hashing or boundary comparison
+THREAD_MEMORY_BUFFER = 128 * 1024
+
+# size of beginning and edge boundary chunks to compare
+BOUNDARY_CHECK_SIZE = 4096
 CHECKPOINT_SCHEMA = 'pydupes-v1'
 
 lock = threading.Lock()
 logger = logging.getLogger('pydupes')
+thread_local = threading.local()
 
 
 def none_if_io_error(f):
@@ -43,13 +50,19 @@ def sizeof_fmt(num, suffix="B"):
 
 @none_if_io_error
 def sha256sum(path: str) -> bytes:
-    h = hashlib.sha256()
-    b = bytearray(128 * 1024)
-    mv = memoryview(b)
+    hasher = hashlib.sha256()
+    view = thread_local.memory_view
     with open(path, 'rb', buffering=0) as f:
-        for n in iter(lambda: f.readinto(mv), 0):
-            h.update(mv[:n])
-    return h.digest()
+        for n in iter(lambda: f.readinto(view), 0):
+            hasher.update(view[:n])
+    return hasher.digest()
+
+
+def _thread_init():
+    thread_local.memory_view = memoryview(bytearray(THREAD_MEMORY_BUFFER))
+
+
+_thread_init()
 
 
 class FatalCrawlException(Exception):
@@ -88,14 +101,15 @@ class DuplicateComparator:
 
 
 class FutureFreeThreadPool:
-
     # TODO: benchmark https://github.com/brmmm3/fastthreadpool
-    def __init__(self, threads: int, queue_size: int = 32):
+    def __init__(self, threads: int, queue_size: int = 32, init_fn=_thread_init):
         assert threads >= 0
         if threads <= 1:
             threads = 0
 
+        self.init_fn = init_fn
         self.work_queue = queue.Queue(maxsize=queue_size)
+        self.queue_size = queue_size
         self.threads = [threading.Thread(target=self._worker_run, daemon=True)
                         for _ in range(threads)]
         self._threadid_to_event = defaultdict(threading.Event)
@@ -116,19 +130,20 @@ class FutureFreeThreadPool:
             t.join()
 
     @staticmethod
-    def _run_task(fn_args):
-        fn, callback, args = fn_args
+    def _run_task(fn, callback, args):
         result = fn(*args)
         if callback:
             callback(result)
 
     def _worker_run(self):
+        if self.init_fn:
+            self.init_fn()
         try:
             while True:
                 fn_args = self.work_queue.get(block=True)
                 if fn_args is None:
                     return
-                self._run_task(fn_args)
+                self._run_task(*fn_args)
                 self.work_queue.task_done()
                 # allow GC
                 del fn_args
@@ -137,8 +152,8 @@ class FutureFreeThreadPool:
             exit(2)
 
     def submit(self, fn, *args, callback=None):
-        if not self.threads or len(self.work_queue.queue) > 1024:
-            self._run_task((fn, callback, args))
+        if not self.threads or len(self.work_queue.queue) >= self.queue_size:
+            self._run_task(fn, callback, args)
         else:
             self.work_queue.put((fn, callback, args))
 
@@ -185,7 +200,8 @@ class FileCrawler:
             raise FatalCrawlException("All roots required to be directories")
         root_stats = [r.lstat() for r in roots]
         root_device_ids = [r.st_dev for r in root_stats]
-        assert len(set(root_device_ids)) == 1, 'Unable to span multiple devices'
+        if len(set(root_device_ids)) != 1:
+            raise FatalCrawlException("Unable to span multiple devices")
         self._root_device = root_device_ids[0]
 
         self._size_to_paths = defaultdict(list)
@@ -197,38 +213,39 @@ class FileCrawler:
         self.num_directories = len(roots)
 
     def _traverse_path(self, path: str):
-        next_dir = None
-        for p in os.scandir(path):
-            ppath = p.path
-            if self._prog is not None:
-                self._prog.update(1)
-            try:
-                pstat = p.stat(follow_symlinks=False)
-            except OSError as e:
-                logger.error('Unable to stat %s due to: %s', ppath, e)
-                continue
-            if pstat.st_dev != self._root_device:
-                logger.debug('Skipping file on separate device %s', ppath)
-                continue
-            if stat.S_ISLNK(pstat.st_mode):
-                logger.debug('Skipping symlink %s', ppath)
-                continue
-            if self._traversed_inodes(pstat.st_ino):
-                logger.debug('Skipping hardlink or already traversed %s', ppath)
-                continue
-            if stat.S_ISREG(pstat.st_mode):
-                if self._min_size <= pstat.st_size:
-                    self._size_to_paths[pstat.st_size].append(ppath)
-            elif stat.S_ISDIR(pstat.st_mode):
-                self.num_directories += 1
-                if not next_dir:
-                    next_dir = ppath
+        frontier = collections.deque()
+        frontier.append(path)
+
+        while frontier:
+            path = frontier.popleft()
+            for p in os.scandir(path):
+                ppath = p.path
+                if self._prog is not None:
+                    self._prog.update(1)
+                try:
+                    pstat = p.stat(follow_symlinks=False)
+                except OSError as e:
+                    logger.error('Unable to stat %s due to: %s', ppath, e)
+                    continue
+                if pstat.st_dev != self._root_device:
+                    logger.debug('Skipping file on separate device %s', ppath)
+                    continue
+                if stat.S_ISLNK(pstat.st_mode):
+                    logger.debug('Skipping symlink %s', ppath)
+                    continue
+                if self._traversed_inodes(pstat.st_ino):
+                    logger.debug('Skipping hardlink or already traversed %s', ppath)
+                    continue
+                if stat.S_ISREG(pstat.st_mode):
+                    if self._min_size <= pstat.st_size:
+                        self._size_to_paths[pstat.st_size].append(ppath)
+                elif stat.S_ISDIR(pstat.st_mode):
+                    self.num_directories += 1
+                    if len(frontier) > 32:
+                        self._pool.submit(self._traverse_path, frontier.popleft())
+                    frontier.append(ppath)
                 else:
-                    self._pool.submit(self._traverse_path, ppath)
-            else:
-                logger.debug('Skipping device/socket/unknown: %s', ppath)
-        if next_dir:
-            self._traverse_path(next_dir)
+                    logger.debug('Skipping device/socket/unknown: %s', ppath)
 
     def traverse(self):
         for r in sorted(set(self._roots)):
@@ -268,10 +285,10 @@ class DupeFinder:
 
     def find(self, size_bytes: int, paths: typing.List[str]) -> typing.Tuple[str, ...]:
         files_start = len(paths)
-        if size_bytes > 8192:
-            path_groups = list(self._split_by_boundaries(size_bytes, paths))
+        if size_bytes > BOUNDARY_CHECK_SIZE * 2:
+            path_groups = tuple(self._split_by_boundaries(size_bytes, paths))
         else:
-            path_groups = [paths]
+            path_groups = (paths,)
         files_end = sum(len(p) for p in path_groups)
         if self._file_progress is not None:
             self._file_progress.update(files_start - files_end)
@@ -304,22 +321,21 @@ class DupeFinder:
 
     @staticmethod
     @none_if_io_error
-    def _read_boundary(path: str, size: int, use_hash=True) -> typing.Union[bytes, int]:
+    def _read_boundary_hash(path: str, size: int) -> int:
         with open(path, 'rb', buffering=0) as f:
             if size < 0:
                 f.seek(size, os.SEEK_END)
             boundary = f.read(abs(size))
-        if use_hash:
             return hash(boundary)
-        return boundary
 
     def _split_by_boundary(self, size_bytes: int, group: typing.List[str], end: bool):
         matches = defaultdict(list)
-        size = -4096 if end else 4096
-        boundaries = self._map(functools.partial(self._read_boundary, size=size), group, size_bytes)
+        size = -BOUNDARY_CHECK_SIZE if end else BOUNDARY_CHECK_SIZE
+        boundaries = self._map(functools.partial(self._read_boundary_hash, size=size), group, size_bytes)
         for path, edge in zip(group, boundaries):
             matches[edge].append(path)
-        return [g for boundary, g in matches.items() if boundary is not None and len(g) > 1]
+        return [paths for boundary, paths in matches.items()
+                if boundary is not None and len(paths) > 1]
 
     def _split_by_boundaries(self, size_bytes: int, group: typing.List[str]):
         groups = self._split_by_boundary(size_bytes, group, end=False)
@@ -342,7 +358,7 @@ class DupeFinder:
 @click.option('--progress', is_flag=True, help='Enable progress bars.')
 @click.option('--min-size', type=click.IntRange(min=0), default=1,
               help='Minimum file size (in bytes) to consider during traversal.')
-@click.option('--read-concurrency', type=click.IntRange(min=1), default=8,
+@click.option('--read-concurrency', type=click.IntRange(min=1), default=4,
               help='I/O concurrency for reading files.')
 @click.option('--traversal-concurrency', type=click.IntRange(min=1), default=1,
               help='I/O concurrency for traversal (stat and listing syscalls).')
